#ifndef _MATRIX_H_
#define _MATRIX_H_

#include "Vertex.h"
#include "Plane.h"

/*
	* The Matrix structure defines a 4x4 Matrix
*/

struct Matrix {
	float m[16];

	Matrix(){}
	
	Matrix(float m11, float m12, float m13, float m14,
		   float m21, float m22, float m23, float m24,
		   float m31, float m32, float m33, float m34,
		   float m41, float m42, float m43, float m44){

		m[0]  = m11; m[1]  = m21; m[2]  = m31; m[3]  = m41;
		m[4]  = m12; m[5]  = m22; m[6]  = m32; m[7]  = m42;
		m[8]  = m13; m[9]  = m23; m[10] = m33; m[11] = m43;
		m[12] = m14; m[13] = m24; m[14] = m34; m[15] = m44;
	}

	Matrix(float initVal){
		for (unsigned int i = 0; i < 16; i++){
			m[i] = initVal;
		}
	}
	
	Matrix(float im[]){
		for (unsigned int i = 0; i < 16; i++){
			m[i] = im[i];
		}
	}

	void loadShadowMatrix(const Plane &plane, const Vertex &lightPos){
		float dist = plane.distance(lightPos);
		
		m[ 0] = dist - lightPos.x * plane.normal.x;
		m[ 4] =      - lightPos.x * plane.normal.y;
		m[ 8] =      - lightPos.x * plane.normal.z;
		m[12] =      - lightPos.x * plane.offset;
		
		m[ 1] =      - lightPos.y * plane.normal.x;
		m[ 5] = dist - lightPos.y * plane.normal.y;
		m[ 9] =      - lightPos.y * plane.normal.z;
		m[13] =      - lightPos.y * plane.offset;
		
		m[ 2] =      - lightPos.z * plane.normal.x;
		m[ 6] =      - lightPos.z * plane.normal.y;
		m[10] = dist - lightPos.z * plane.normal.z;
		m[14] =      - lightPos.z * plane.offset;
		
		m[ 3] =      - plane.normal.x;
		m[ 7] =      - plane.normal.y;
		m[11] =      - plane.normal.z;
		m[15] = dist - plane.offset;
	}


	void loadIdentity(){
		for (unsigned int i = 0; i < 16; i++){
			m[i] = ((i & 3) == (i >> 2));
		}
	}

	void translate(const Vertex &v){
		m[12] += (v.x * m[0] + v.y * m[4] + v.z * m[8]);
		m[13] += (v.x * m[1] + v.y * m[5] + v.z * m[9]);
		m[14] += (v.x * m[2] + v.y * m[6] + v.z * m[10]);
		m[15] += (v.x * m[3] + v.y * m[7] + v.z * m[11]);
	}

	void scale(const Vertex &v){
		m[0] *= v.x; m[1] *= v.x; m[2]  *= v.x; m[3]  *= v.x;
		m[4] *= v.y; m[5] *= v.y; m[6]  *= v.y; m[7]  *= v.y;
		m[8] *= v.z; m[9] *= v.z; m[10] *= v.z; m[11] *= v.z;
	}
/*
	void rotateX(const float angle){
	}
	void rotateY(const float angle){
	}
	void rotateZ(const float angle){
	}
*/

	void rotate(const Vertex &v, float angle){
		Matrix temp;
		float cosA = cosf(angle),
			  sinA = sinf(angle);

		temp.m[0]  = v.x * v.x * (1 - cosA) + cosA;
		temp.m[1]  = v.x * v.y * (1 - cosA) + v.z * sinA;
		temp.m[2]  = v.x * v.z * (1 - cosA) - v.y * sinA;
		temp.m[3]  = 0;
		temp.m[4]  = v.x * v.y * (1 - cosA) - v.z * sinA;
		temp.m[5]  = v.y * v.y * (1 - cosA) + cosA;
		temp.m[6]  = v.y * v.z * (1 - cosA) + v.x * sinA;
		temp.m[7]  = 0;
		temp.m[8]  = v.x * v.z * (1 - cosA) + v.y * sinA;
		temp.m[9]  = v.y * v.z * (1 - cosA) - v.x * sinA;
		temp.m[10] = v.z * v.z * (1 - cosA) + cosA;
		temp.m[11] = 0;
		temp.m[12] = 0;
		temp.m[13] = 0;
		temp.m[14] = 0;
		temp.m[15] = 1;

		(*this) *= temp;
	}

	void transpose(){
		float temp;

		temp  = m[4];
		m[4]  = m[1];
		m[1]  = temp;

		temp  = m[8];
		m[8]  = m[2];
		m[2]  = temp;

		temp  = m[12];
		m[12] = m[3];
		m[3]  = temp;

		temp  = m[9];
		m[9]  = m[6];
		m[6]  = temp;

		temp  = m[13];
		m[13] = m[7];
		m[7]  = temp;

		temp  = m[14];
		m[14] = m[11];
		m[11] = temp;
	}

	void print(){
		printf("%f %f %f %f\n",   m[0], m[4], m[8], m[12]);
		printf("%f %f %f %f\n",   m[1], m[5], m[9], m[13]);
		printf("%f %f %f %f\n",   m[2], m[6], m[10],m[14]);
		printf("%f %f %f %f\n\n", m[3], m[7], m[11],m[15]);
	}

	const float &operator () (const int row, const int column);
	const float &operator [] (const int index);
	void operator += (const Matrix &v);
	void operator -= (const Matrix &v);
	void operator *= (const Matrix &v);

	void operator += (const float term);
	void operator -= (const float term);
	void operator *= (const float scalar);
	void operator /= (const float dividend);
};

const float &Matrix::operator () (const int row, const int column){
	return m[row + (column << 2)];
}

const float &Matrix::operator [] (const int index){
	return m[index];
}

Matrix operator + (const Matrix &u, const Matrix &v){
	Matrix mat;
	for (unsigned int i = 0; i < 16; i++){
		mat.m[i] = u.m[i] + v.m[i];
	}
	return mat;
}

Matrix operator - (const Matrix &u, const Matrix &v){
	Matrix mat;
	for (unsigned int i = 0; i < 16; i++){
		mat.m[i] = u.m[i] - v.m[i];
	}
	return mat;
}

Matrix operator - (const Matrix &v){
	Matrix mat;
	for (unsigned int i = 0; i < 16; i++){
		mat.m[i] = -v.m[i];
	}
	return mat;
}

void Matrix::operator += (const Matrix &v){
	for (unsigned int i = 0; i < 16; i++){
		m[i] += v.m[i];
	}
}

void Matrix::operator -= (const Matrix &v){
	for (unsigned int i = 0; i < 16; i++){
		m[i] -= v.m[i];
	}
}


void Matrix::operator += (const float term){
	for (unsigned int i = 0; i < 16; i++){
		m[i] += term;
	}
}

void Matrix::operator -= (const float term){
	for (unsigned int i = 0; i < 16; i++){
		m[i] -= term;
	}
}

void Matrix::operator *= (const float scalar){
	for (unsigned int i = 0; i < 16; i++){
		m[i] *= scalar;
	}
}

void Matrix::operator /= (const float dividend){
	for (unsigned int i = 0; i < 16; i++){
		m[i] /= dividend;
	}
}

Matrix operator * (const Matrix &u, const Matrix &v){
	Matrix mat;
	unsigned int i,us,vs;

	for (i = 0; i < 16; i++){
		us = (i & 0x3);
		vs = (i & 0xC);

		mat.m[i] =  u.m[us     ] * v.m[vs    ] + 
					u.m[us +  4] * v.m[vs + 1] + 
					u.m[us +  8] * v.m[vs + 2] + 
					u.m[us + 12] * v.m[vs + 3];
	}
	return mat;
}

void Matrix::operator *= (const Matrix &v){
	unsigned int i,us,vs;
	float nm[16];

	for (i = 0; i < 16; i++){
		us = (i & 0x3);
		vs = (i & 0xC);

		nm[i] = m[us     ] * v.m[vs    ] + 
				m[us +  4] * v.m[vs + 1] + 
				m[us +  8] * v.m[vs + 2] + 
				m[us + 12] * v.m[vs + 3];
	}
	for (i = 0; i < 16; i++){
		m[i] = nm[i];
	}
}

Vertex operator * (const Matrix &mat, const Vertex &v){
	return 
		Vertex( mat.m[0] * v.x + mat.m[4] * v.y + mat.m[8]  * v.z, 
				mat.m[1] * v.x + mat.m[5] * v.y + mat.m[9]  * v.z, 
				mat.m[2] * v.x + mat.m[6] * v.y + mat.m[10] * v.z  );
}


Matrix operator + (const Matrix &v, const float term){
	Matrix mat;
	unsigned int i;

	for (i = 0; i < 16; i++){
		mat.m[i] = v.m[i] + term;
	}
	return mat;
}

Matrix operator - (const Matrix &v, const float term){
	Matrix mat;
	unsigned int i;

	for (i = 0; i < 16; i++){
		mat.m[i] = v.m[i] - term;
	}
	return mat;
}

Matrix operator * (const float scalar, const Matrix &v){
	Matrix mat;
	unsigned int i;

	for (i = 0; i < 16; i++){
		mat.m[i] = v.m[i] * scalar;
	}
	return mat;
}

Matrix operator * (const Matrix &v, const float scalar){
	Matrix mat;
	unsigned int i;

	for (i = 0; i < 16; i++){
		mat.m[i] = v.m[i] * scalar;
	}
	return mat;
}

Matrix operator / (const Matrix &v, const float dividend){
	Matrix mat;
	unsigned int i;

	for (i = 0; i < 16; i++){
		mat.m[i] = v.m[i] / dividend;
	}
	return mat;
}

Matrix operator ! (const Matrix &v){
	Matrix mat;
	unsigned int i,j,k;
	float temp, newMat[4][8];

	for (k = i = 0; i < 4; i++){
		for (j = 0; j < 4; j++){
			newMat[j][i] = v.m[k];
			newMat[i][j + 4] = (i == j);
			k++;
		}
	}


	for (i = 0; i < 4; i++){
		// Flip rows if neccesary. It doesn't take care of uninvertible matrixes.
		if (newMat[i][i] == 0){
			j = 0;
			do {
				j++;
			} while (newMat[j][i] == 0);

			for (k = i; k < 8; k++){
				temp = newMat[j][k];
				newMat[j][k] = newMat[i][k];
				newMat[i][k] = temp;
			}
		}
		
		// Shorten the numbers the current row to get a 1 at (i,i) ... the actual 1 is uninteresting and omitted
		for (j = i+1; j < 8; j++){
			newMat[i][j] /= newMat[i][i];
		}

		// Get zeros in the beginning ... the actual zeros are uninteresting and omitted
		for (j = 0; j < 4; j++){
			if (j != i){
				for (k = i+1; k < 8; k++){
					newMat[j][k] -= newMat[j][i] * newMat[i][k];
				}
			}
		}
	}

	// Pack it into the return Matrix object.
	for (k = 0, i = 4; i < 8; i++){
		for (j = 0; j < 4; j++){
			mat.m[k] = newMat[j][i];
			k++;
		}
	}

	return mat;
}

#endif